unit TermInterpreter;

interface

function TermWert(s: String; x: Double): Double;
function IsOkay: Boolean;
function GetLastErrorMsg: String;
function GetLastErrorPos: Integer;

implementation

uses math;

const fnmax = 12;
      fname : Array[1..fnmax] of String =
               ('SQR', 'SQRT', 'LOG', 'LN', 'EXP',
                'SIN', 'COS', 'TAN', 'ARCTAN', 'ABS',
                'SGN', 'RND');

var LastErrorMsg : String;
    LastErrorPos : Integer;
    Okay         : Boolean;
    const_e,
    arg_x        : Double;


{ Interne Hilfsfunktionen }

function GetFirstChar(var s: String): Char;
  begin
  While (Length(s) > 0) and (s[1] in [' ']) do
    Delete(s, 1, 1);             { killt fhrende Leerzeichen        }
  If Length(s) > 0 then
    Result := s[1]               { gibt das erste Zeichen zurck.... }
  else
    Result := #0;                { ... oder bei leerem String ein #0 }
  end;

function signum(r: Double): Double;
  begin
  If r > 0 then Result := 1
  else if r < 0 then Result := -1
  else Result := 0;
  end;

function ClosingBracket(c: Char): Char;
  begin
  Case c of
    '(' : Result := ')';
    '[' : Result := ']';
    '{' : Result := '}';
  else
    Result := #0;
  end; {of case }
  end;

function GetFirstDenominator(s: String): String;
  var i : Integer;
  begin
  While (Length(s) > 0) and (s[1] in [' ']) do
    Delete(s, 1, 1);             { killt fhrende Leerzeichen        }
  i := 1;
  Result := '';
  While (i <= Length(s)) and (s[i] in ['A'..'Z', 'a'..'z']) do begin
    Result := Result + Upcase(s[i]);
    i := i + 1;
    end;
  end;

procedure DeleteFirstDenominator(var s: String);
  var i : Integer;
  begin
  i := 1;
  While (i <= Length(s)) and (s[i] in ['A'..'Z', 'a'..'z']) do
    i := i + 1;
  Delete(s, 1, i-1);
  end;

function GetFunctionNumber(fn: String): Integer;
  var i : Integer;
  begin
  Result := -1;    { Funktionsname nicht in der Liste gefunden }
  i := 1;
  While (Result < 0) and (i <= fnmax) do
    If fn = fname[i] then
      Result := i
    else
      i := i + 1;
  end;


{ Scannerfunktionen fr logische Term-Bestandteile }

function N_Zahl(var s: String; var wert: Double): Boolean;
  var c : Char;
  begin
  c := GetFirstChar(s);
  If c in ['0'..'9'] then begin  { Es wird eine Ziffer erwartet....    }
    Result := True;
    wert := 0;
    Repeat                       { Solange Ziffern da sind, werden sie }
      Delete(s, 1, 1);
      wert := 10 * wert + (Ord(c) - Ord('0'));  { ... ausgewertet.       }
      c := GetFirstChar(s);                   { Nchstes Zeichen holen }
    until not (c in ['0'..'9']); { bis keine Ziffer mehr gefunden wird }
    end
  else begin                     { Falls keine Ziffer da ist: Fehler!  }
    Result := False;
    LastErrorMsg := 'Ziffer erwartet!';       { Fehlermeldung absetzen }
    end;
  end;


function Zahl(var s: String; var wert: Double): Boolean;
  var IsMinus,
      IsMinusExp : Boolean;
      c          : Char;
      d, e       : Double;
  begin
  IsMinus := False;              { Normalerweise erwarten wir was Positives! }
  c := GetFirstChar(s);
  If c in ['+', '-'] then begin        { Optionales Vorzeichen testen }
    Delete(s, 1, 1);
    IsMinus := c = '-';                { Gegebenenfalls auf Minus umschalten }
    end;
  If N_Zahl(s, wert) then begin         { Vorkomma-Anteil holen }
    Result := True;                    { => Erfolg ! }
    c := GetFirstChar(s);
    If c in ['.', ','] then begin      { Nachkommastellen vorhanden ? }
      Delete(s, 1, 1);                 { Dezimaltrenner lschen }
      c := GetFirstChar(s);
      If c in ['0'..'9'] then begin    { Ziffern einlesen }
        d := 0.1;                      { Stellenwert initialisieren }
        Repeat
          Delete(s, 1, 1);
          wert := wert + d * (Ord(c) - Ord('0'));  { Wert berechnen }
          d   := d / 10;               { Stellenwert fr nchste Stelle }
          c   := GetFirstChar(s);      { Nchste Ziffer holen ... }
        until not (c in ['0'..'9']);   { ...bis keine mehr da sind }
        end
      else begin
        Result := False;
        LastErrorMsg := 'Nachkommastellen erwartet!';
        end;
      end;
    If c in ['e', 'E'] then begin      { Exponent vorhanden ? }
      Delete(s, 1, 1);                 { 'e' lschen }
      IsMinusExp := False;
      c := GetFirstChar(s);
      If c in ['+', '-'] then begin    { optionales Vorzeichen testen }
        Delete(s, 1, 1);
        IsMinusExp := c = '-';         { gegebenenfalls auf Minus umschalten }
        end;
      If N_Zahl(s, e) then begin       { Exponent holen }
        If IsMinusExp then e := -e;
        wert := wert * power(10, e);
        end
      else begin                       { Kein Exponent vorhanden ? => Fehler! }
        Result := False;
        LastErrorMsg := 'Exponent erwartet!';
        end;
      end;
    If IsMinus then wert := -wert;       { Erst hier das VZ einrechnen !!!      }
    end
  else
    Result := False;            { LastErrorMsg wird schon von N_Zahl gesetzt. }
  end;


function Summe(var s: String; var wert: Double): Boolean; forward;
  { Ntig, weil in der folgenden Funktion "Klammerterm" die erst
    weiter unten deklarierte Funktion "Summe" aufgerufen wird.   }


function Klammerterm(var s: String; var wert: Double): Boolean;
  var c1, c2 : Char;
  begin
  c1 := GetFirstChar(s);
  If c1 in ['(', '[', '{'] then begin  { Klammerterm }
    Delete(s, 1, 1);
    If Summe(s, wert) then begin        { Klammerinhalt berechnen }
      c2 := GetFirstChar(s);
      If c2 = ClosingBracket(c1) then begin
        Delete(s, 1, 1);
        Result := True;
        end
      else begin
        Result := False;
        LastErrorMsg := '"' + ClosingBracket(c1)+ '" erwartet!';
        end;
      end
    else
      Result := False;                 { Fehlermeldung bernehmen }
    end
  else
    Result := False;
  end;


function Funktion(var s: String; var wert: Double): Boolean;
  var fname : String;
      fnum  : Integer;
      arg   : Double;
  begin
  Result := False;      { Pessimistische Annahme, die Quelltext sparen hilft }
  fname  := GetFirstDenominator(s);
  fnum   := GetFunctionNumber(fname);
  If fnum >= 0 then begin
    DeleteFirstDenominator(s);
    If Klammerterm(s, arg) then
      Case fnum of
        1 : begin                  // SQR
            wert := sqr(arg);
            Result := True;
            end;
        2 : If arg >= 0 then begin // SQRT
              wert := sqrt(arg);
              Result := True;
              end;
        3 : If arg > 0 then begin  // LOG
              wert := log10(arg);
              Result := True;
              end;
        4 : If arg > 0 then begin  // LN
              wert := ln(arg);
              Result := True;
              end;
        5 : begin                  // EXP
            wert := exp(arg);
            Result := True;
            end;
        6 : begin                  // SIN
            wert := sin(arg);
            Result := True;
            end;
        7 : begin                  // COS
            wert := cos(arg);
            Result := True;
            end;
        8 : begin                  // TAN
            wert := tan(arg);
            Result := True;
            end;
        9 : begin                  // ARCTAN
            wert := arctan(arg);
            Result := True;
            end;
       10 : begin                  // ABS
            wert := abs(arg);
            Result := True;
            end;
       11 : begin
            wert := signum(arg);    // SGN
            Result := True;
            end;
       12 : begin
            wert := ROUND(arg);     // RND
            Result := True;
            end;
      end; { of case }
    end;
  end;


function VarKonst(var s: String; var wert: Double): Boolean;
  var c1, c2 : Char;
  begin
  Result := True;     { Optimistische Annahme, die Quelltext sparen hilft }
  c1 := Upcase(GetFirstChar(s));
  Case c1 of
    'X' : begin
          Delete(s, 1, 1);
          wert := arg_x;
          end;
    'E' : begin
          Delete(s, 1, 1);
          wert := const_e;
          end;
    'P' : begin
          Delete(s, 1, 1);
          c2 := Upcase(GetFirstChar(s));
          If c2 = 'I' then begin
            Delete(s, 1, 1);
            wert := Pi;
            end
          else
            Result := False;
          end;
  else
    Result := False;
  end; { of case }
  end;


function VFaktor(var s: String; var wert: Double; var HasMinus: Boolean): Boolean;
  var c : Char;
  begin
  HasMinus := False;              { Normalerweise erwarten wir was Positives! }
  c := GetFirstChar(s);
  If c in ['+', '-'] then begin      { Optionales Vorzeichen testen         }
    Delete(s, 1, 1);
    HasMinus := c = '-';             { Gegebenenfalls auf Minus umschalten  }
    end;
  Result := Klammerterm(s, wert) or   { Jede dieser Funktionen wird nur dann }
            Zahl(s, wert) or          { wirklich aufgerufen, wenn alle zuvor }
            Funktion(s, wert) or      { aufgerufenen das Ergebnis FALSE      }
            VarKonst(s, wert);        { zurckgegeben haben.                 }
  If HasMinus then                    { Vorzeichen erst hier einrechnen !    }
    wert := -wert;
  If Result then
    LastErrorMsg := '';              { Eventuelle Fehlermeldung lschen     }
  end;


function Potenz(var s: String; var wert: Double): Boolean;
  var BaseHasMinus,
      ExpHasMinus : Boolean;
      r1 : Double;
      c  : Char;
  begin
  If VFaktor(s, wert, BaseHasMinus) then begin     { Basis holen }
    Result := True;
    c := GetFirstChar(s);
    While c = '^' do begin       { Falls Exponent vorhanden ist: }
      Delete(s, 1, 1);             { Hochzeichen lschen und...  }
      If VFaktor(s, r1, ExpHasMinus) then begin { Exponent holen }
        If BaseHasMinus then
          wert := - power(-wert, r1)  { Damit -3^2 < 0 wird ! }
        else
          wert := power(wert, r1);
        c := GetFirstChar(s);
        end
      else begin
        Result := False;          { Fehlermeldung bernehmen    }
        c := #0;
        end;
      end;
    end                         { Kein Exponent => nix zu tun ! }
  else
    Result := False;            { Fehlermeldung bernehmen      }
  end;


function Produkt(var s: String; var wert: Double): Boolean;
  var r1 : Double;
      c  : Char;
  begin
  If Potenz(s, wert) then begin     { Ersten Faktor holen                     }
    Result := True;
    c := GetFirstChar(s);
    While c in ['*', '/'] do begin { Solange eine Punktrechnung folgt :      }
      Delete(s, 1, 1);                 { Rechenzeichen lschen               }
      If Potenz(s, r1) then begin      { Nchsten Faktor holen...            }
        If c = '*' then
          wert := wert * r1              { und zum Produkt multiplizieren .... }
        else
          wert := wert / r1;             { ...bzw. dividieren                  }
        c := GetFirstChar(s);
        end
      else begin                       { Kein Faktor vorhanden ? => Fehler!  }
        Result := False;               {  LastErrorMsg wird in Zahl gesetzt  }
        c      := #0;                  { While-Schleife verlassen }
        end;
      end;
    end
  else                             { Kein Faktor da ? => Fehler!             }
    Result := False;               {     LastErrorMsg wird in Zahl gesetzt   }
  end;


function Summe(var s: String; var wert: Double): Boolean;
  var r1 : Double;
      c  : Char;
  begin
  If Produkt(s, wert) then begin    { Ersten Summanden holen                 }
    Result := True;
    c := GetFirstChar(s);
    While c in ['+', '-'] do begin { Solange eine Strichrechnung folgt :    }
      Delete(s, 1, 1);             { Rechenzeichen lschen                  }
      If Produkt(s, r1) then begin { Nchsten Summanden holen...            }
        If c = '+' then
          wert := wert + r1          { und zur aktuellen Summe addieren...    }
        else
          wert := wert - r1;         { ...bzw. von ihr subtrahieren           }
        c := GetFirstChar(s);
        end
      else begin                   { Kein Summand vorhanden ? => Fehler!    }
        Result := False;           {  LastErrorMsg wird in Produkt gesetzt  }
        c      := #0;              { While-Schleife verlassen }
        end;
      end;
    end
  else                             { Kein Summand da ? => Fehler!           }
    Result := False;               {  LastErrorMsg wird in Produkt gesetzt  }
  end;


{ Interface-Funktionen (ffentlich) }

function TermWert(s: String; x: Double): Double;
  { Diese Funktion dient hauptschlich der Fehlerverwaltung. Sie sorgt dafr,
    dass bei einem Fehler whrend des Interpretationsvorgangs die Variablen
    LastErrorMsg und LastErrorPos korrekt gesetzt werden. Auerdem verwaltet
    sie die Variable Okay, in der ber den Erfolg des Interpretationsversuchs
    Buch gefhrt wird.
    Im Gegensatz zu den internen Interpreter-Funktionen  }
  var r   : Double;
      ils : Integer;
  begin
  Okay  := True;
  arg_x := x;
  ils   := Length(s);
  LastErrorMsg := '';
  LastErrorPos := ils;
  IF Summe(s, r) then begin  { Hier wird gerechnet: jeder Term ist eine Summe. }
    Result := r;
    Okay   := True;
    end
  else begin                    { Buchhaltung im Fehlerfall: }
    Result := 0;
    Okay := False;
    LastErrorPos := ils - Length(s);
    LastErrorMsg := LastErrorMsg + '  {Rest : ' + s + ' }';
    end;
  end;

function IsOkay: Boolean;
  begin
  Result := Okay;
  end;

function GetLastErrorMsg: String;
  begin
  Result := LastErrorMsg;
  end;

function GetLastErrorPos: Integer;
  begin
  Result := LastErrorPos;
  end;

initialization
  LastErrorMsg := '';
  Okay     := True;
  const_e  := exp(1);
end.
